/* flownode.c
   05.05.2004 */

#include <stdlib.h>
#include "flownode.h"


/* functions and procedures */


/*
  FlowGraph generateFlowGraph(ParseTree tree)
  generates a FlowGraph structure from the given
  ParseTree
  Returns:
    NULL when an error occured
    Pointer to an FlowGraph_ structure 
*/
FlowGraph generateFlowGraph(ParseTree tree)
{
    ParseTreeCell walker;
    FlowGraph buffer, result;
    if (tree == NULL) return NULL;
    switch (tree->type)
    {
	case ASSIGN:
	    return simpleFlowGraph(tree->label, FLOW_ASSIGN, tree);
	case SKIP:
	    return simpleFlowGraph(tree->label, FLOW_SKIP, tree);
	case IF:
	    return ifFlowGraph(generateFlowGraph(getSubTree(tree, 1)),
			       generateFlowGraph(getSubTree(tree, 2)), 
						 tree->label,
						 FLOW_IF,
						 tree);
	case SEQ:
	    result = sequenceFlowGraph(generateFlowGraph(getSubTree(tree, 0)),
				       generateFlowGraph(getSubTree(tree, 1)));
	    walker = tree->tag.subtrees->first->next->next;
	    while (walker != NULL)
	    {
		result = sequenceFlowGraph(result,
					   generateFlowGraph(walker->value));
		walker = walker->next;
	    }
	    return result;
    
    case WHILE:
      return whileFlowGraph(generateFlowGraph(getSubTree(tree, 1)), tree->label, FLOW_WHILE, tree);
    default: printf("ERROR Wrong ParseTree; kein GET/PUT in zu analysierenden Programmen nutzen!\n");

    }
}


/*
  FlowNode createFlowNode(int label, int type, ParseTree tree)
  creates a new FlowNode with the given label, type and sub tree
  When an error occured, NULL will be returned
*/
FlowNode createFlowNode(int label, int type, ParseTree tree)
{
    FlowNode result;
    if ((result = (FlowNode)malloc(sizeof(struct FlowNode_))) == NULL)
	return NULL; /* error: no memory allocated */
    result->expr = tree;
    result->label = label;
    result->type = type;
    result->entry = NULL;
    result->exit = NULL;
    return result;
}


/*
  FlowGraph simpleFlowGraph(int label, int type, ParseTree tree)
  creates a "simple" FlowGraph, which consists of only one FlowNode,
  which is entry and exit
*/
FlowGraph simpleFlowGraph(int label, int type, ParseTree tree)
{
    FlowGraph result;
    FlowNode simpleNode;
    if ((result = (FlowGraph)malloc(sizeof(struct FlowGraph_))) == NULL)
	return NULL;
    simpleNode = createFlowNode(label, type, tree);
    simpleNode->entry = NULL;
    simpleNode->exit = NULL;
    result->entry = simpleNode;
    if ((result->exit = (FlowNodeList)malloc(sizeof(struct FlowNodeList_))) == NULL)
	return NULL;
    result->exit->next = NULL;
    result->exit->node = simpleNode;
    result->all = (FlowNodeList)malloc(sizeof(struct FlowNodeList_));
    if (result->all == NULL) return NULL;
    result->all->next = NULL;
    result->all->node = simpleNode;
    return result;
}


/*
  FlowGraph sequenceFlowGraph(FlowGraph a, FlowGraph b)
  creates a FlowGraph which is induced by a sequence of commands ("a;b")
*/
FlowGraph sequenceFlowGraph(FlowGraph a, FlowGraph b)
{
  FlowNodeList walker, combiner;
  FlowNodeList buffer;
  FlowGraph result;
  
  /* initalize the result graph */
  if ((result = (FlowGraph)malloc(sizeof(struct FlowGraph_))) == NULL)
      return NULL; /* error: no memory allocated */
  result->entry = a->entry;
  result->exit = b->exit;  
    /*
      forall k outof a->exit do
         k->exit := k->exit combine b->entry
       od;
     */
    walker = a->exit;
    while (walker != NULL)
    {
	if ((buffer = (FlowNodeList)malloc(sizeof(struct FlowNodeList_))) == NULL)
	    return NULL; /* error: no memory allocated */
	buffer->node = b->entry;
	buffer->next = walker->node->exit;
	walker->node->exit = buffer;
	if ((buffer = (FlowNodeList)malloc(sizeof(struct FlowNodeList_))) == NULL)
	    return NULL; /* error: no memory allocated */
	buffer->node = walker->node;
	buffer->next = b->entry->entry;
	b->entry->entry = buffer;        
	walker = walker->next;
    }  

    /* append the all list from b to a */
    walker = a->all;
    while (walker->next != NULL)
      walker = walker->next;
    walker->next = b->all;
    result->all = a->all;  
    /* toDo: Lschen der unntigen a,b-Listen (b->entry, a->exit) */
    /* deallocate the obsolete lists and structures */
    deleteFlowNodeList(a->exit);
    free(a);
    free(b);
    return result;
} /* end sequenceFlowGraph */


/*
  FlowGraph ifFlowGraph(FlowGraph thenbranch, FlowGraph elsebranch, int label, int type, ParseTree expr)
  creates a FlowGraph which is induced by an if-statement (if BED then thenbranch else elsebranch fi)
*/
FlowGraph ifFlowGraph(FlowGraph thenbranch, FlowGraph elsebranch, int label, int type, ParseTree expr)
{
   
  FlowGraph result;
  FlowNode ifNode;
  FlowNodeList buffer, walker;
  if (((result = (FlowGraph)malloc(sizeof(struct FlowGraph_))) == NULL)
     ||((ifNode = (FlowNode)malloc(sizeof(struct FlowNode_))) == NULL)
      ||((buffer = (FlowNodeList)malloc(sizeof(struct FlowNodeList_))) == NULL))
    return NULL; /* error: no memory allocated */
  result->entry = ifNode;
   
  /* exits are: exit(then) combine exit(else) */
  /* we generate the "new" exit list by appending the thenbranch list to
     the elsebranch (we don't need them anymore) */
  ifNode->exit = NULL;
  walker = thenbranch->exit;
  
  /* there have to be at least one exit, so we can simply append the two lists */
  while (walker->next != NULL)
      walker = walker->next;
  walker->next = elsebranch->exit;
  result->exit = thenbranch->exit;
  result->entry = ifNode;

  /* setting the values of the ifNode */
  ifNode->label = label;
  ifNode->type = type;
  ifNode->expr = expr;
  ifNode->entry = NULL;

  /* setting the entry-slots of the (old) entry Nodes */
  if ((buffer = (FlowNodeList)malloc(sizeof(struct FlowNodeList_))) == NULL)
    return NULL; /* error: no memory allocated */
  buffer->node = ifNode;
  buffer->next = thenbranch->entry->entry;
  thenbranch->entry->entry = buffer;

  if ((buffer = (FlowNodeList)malloc(sizeof(struct FlowNodeList_))) == NULL)
    return NULL; /* error: no memory allocated */
  buffer->node = ifNode;
  buffer->next = elsebranch->entry->entry;
  elsebranch->entry->entry = buffer;
  
  if ((buffer = (FlowNodeList)malloc(sizeof(struct FlowNodeList_))) == NULL)
    return NULL; /* error: no memory allocated */
  buffer->node = thenbranch->entry;
  buffer->next = ifNode->exit;
  ifNode->exit = buffer;
 
  if ((buffer = (FlowNodeList)malloc(sizeof(struct FlowNodeList_))) == NULL)
    return NULL; /* error: no memory allocated */
  buffer->node = elsebranch->entry;
  buffer->next = ifNode->exit;
  ifNode->exit = buffer;

  /* append the all list from b to a */
  walker = thenbranch->all;
  while (walker->next != NULL)
    walker = walker->next;
  walker->next = elsebranch->all;
  /* finally, insert the ifNode */
  if ((buffer = (FlowNodeList)malloc(sizeof(struct FlowNodeList_))) == NULL)
    return NULL; /* error: no memory allocated */
  buffer->next = thenbranch->all;
  buffer->node = ifNode;
  result->all = buffer;
  free(elsebranch);
  free(thenbranch);
  return result;
}


/*
  FlowGraph whileFlowGraph(FlowGraph whileProg, int label, int type, ParseTree expr)
  creates a FlowGraph which is induced by a while loop; whileProg is the body of the
  loop
*/
FlowGraph whileFlowGraph(FlowGraph whileProg, int label, int type, ParseTree expr)
{
  FlowGraph result;
  FlowNodeList walker, buffer;
  FlowNode whileNode;
  
  whileNode = createFlowNode(label, type, expr);
  whileNode->entry = NULL;
  result = (FlowGraph)malloc(sizeof(struct FlowGraph_));
  if (result == NULL) return NULL;
  
  result->entry = whileNode;
 
  /* there is only one exit node: the whileNode */
  buffer = (FlowNodeList)malloc(sizeof(struct FlowNodeList_));
  if (buffer == NULL) return NULL;
  
  buffer->next = NULL;
  buffer->node = whileNode;
  result->exit = buffer;

  buffer = (FlowNodeList)malloc(sizeof(struct FlowNodeList_));
  if (buffer == NULL) return NULL;
  buffer->node = whileNode;
  buffer->next = whileProg->entry->entry;
  whileProg->entry->entry = buffer;

  buffer = (FlowNodeList)malloc(sizeof(struct FlowNodeList_));
  if (buffer == NULL) return NULL;
  buffer->node = whileProg->entry;
  buffer->next = whileNode->exit;
  whileNode->exit = buffer;

  walker = whileProg->exit;
  
  while (walker != NULL)
    {
      buffer = (FlowNodeList)malloc(sizeof(struct FlowNodeList_));
      if (buffer == NULL) return NULL;
      buffer->node = whileNode;
      buffer->next = walker->node->exit;
      walker->node->exit = buffer;
      buffer = (FlowNodeList)malloc(sizeof(struct FlowNodeList_));
      if (buffer == NULL) return NULL;
      buffer->node = walker->node;
      buffer->next = whileNode->entry;
      whileNode->entry = buffer;
      walker = walker->next;
    }
  
  if ((buffer = (FlowNodeList)malloc(sizeof(struct FlowNodeList_))) == NULL)
    return NULL; /* error: no memory allocated */
  buffer->next = whileProg->all;
  buffer->node = whileNode;
  result->all = buffer;
  deleteFlowNodeList(whileProg->exit);
  free(whileProg);
  return result;
}
