/* flownode.h 

  Implements an ADT FlowGraph and FlowNode,
  which are used to construct the flow graph of the 
  analysis.
  FlowGraph:
    consists of
    - the entry node ("first block")
    - the exit node list ("exit blocks")
    - the all node list ("all used blocks")
  FlowNode:
    consists of
    - the corresponding subtree in the ParseTree (expr)
    - the label id of the block
    - type of the FlowNode (for later use, not used in wAnalyzer)
    - the corresponding analysis result 
    - the entrys / exits FlowNodes
  
  The FlowGraph of a given program is constructed recursively
    
   05.05.2004 
*/
   
#ifndef _FLOWNODE_H
#define _FLOWNODE_H

#include <stdlib.h>
#include "parsetree.h"
#include "symbols.h"
#include "generic_list.h"


#define FLOW_IF IF
#define FLOW_WHILE WHILE
#define FLOW_SKIP SKIP
#define FLOW_ASSIGN ASSIGN

  
typedef struct FlowNodeList_
  {
  struct FlowNodeList_ *next;
  struct FlowNode_ *node;
  } *FlowNodeList;

typedef struct FlowNode_
{
    ParseTree expr;
    int label, type;
    Element analysis;
    FlowNodeList entry, exit;
} *FlowNode;
  

typedef struct FlowGraph_
{
    FlowNode entry;
    FlowNodeList exit;
    FlowNodeList all;
} *FlowGraph;



/* functions and procedures */


/*
  FlowGraph generateFlowGraph(ParseTree tree)
  generates a FlowGraph structure from the given
  ParseTree
  Returns:
    NULL when an error occured
    Pointer to an FlowGraph_ structure 
*/
FlowGraph generateFlowGraph(ParseTree tree);

/*
  FlowNode createFlowNode(int label, int type, ParseTree tree)
  creates a new FlowNode with the given label, type and sub tree
  When an error occured, NULL will be returned
*/
FlowNode createFlowNode(int label, int type, ParseTree tree);

/*
  FlowGraph simpleFlowGraph(int label, int type, ParseTree tree)
  creates a "simple" FlowGraph, which consists of only one FlowNode,
  which is entry and exit
*/
FlowGraph simpleFlowGraph(int label, int type, ParseTree tree);

/*
  FlowGraph whileFlowGraph(FlowGraph whileProg, int label, int type, ParseTree expr)
  creates a FlowGraph which is induced by a while loop; whileProg is the body of the
  loop
*/
FlowGraph whileFlowGraph(FlowGraph whileProg, int label, int type, ParseTree expr);

/*
  FlowGraph sequenceFlowGraph(FlowGraph a, FlowGraph b)
  creates a FlowGraph which is induced by a sequence of commands ("a;b")
*/
FlowGraph sequenceFlowGraph(FlowGraph a, FlowGraph b);

/*
  FlowGraph ifFlowGraph(FlowGraph thenbranch, FlowGraph elsebranch, int label, int type, ParseTree expr)
  creates a FlowGraph which is induced by an if-statement (if BED then thenbranch else elsebranch fi)
*/
FlowGraph ifFlowGraph(FlowGraph thenbranch, FlowGraph elsebranch, int label, int type, ParseTree expr);


#endif 

